from collections import Counter 

def get_character_counts(text):
    """simply use the Counter class from the collections module to count the occurrences of each character"""
    return Counter(text)

# arithmetic encoding
def encode(message, frequencies, cumulative_frequencies):
    # initialize the interval
    low = 0
    high = 1
    # encode the message
    for char in message:
        # calculate the range
        range = high - low
        # calculate the new high and low values
        high = low + range * (cumulative_frequencies[char] + frequencies[char])
        low = low + range * cumulative_frequencies[char]
    return (low + high) / 2

# get the binary code as string from the interval
def get_code_from_interval(interval):
    code = ""
    while interval > 0:
        interval *= 2
        if interval >= 1:
            code += "1"
            interval -= 1
        else:
            code += "0"
    return code

def decode_with_length(encoded_message, frequencies, cumulative_frequencies, length):
    message = ""
    interval = encoded_message
    while len(message) < length:
        for char, freq in frequencies.items():
            if cumulative_frequencies[char] <= interval < cumulative_frequencies[char] + freq:
                message += char
                low = cumulative_frequencies[char]
                high = (cumulative_frequencies[char] + freq)
                interval = (interval - low) / (high - low)
                break
    return message


if __name__ == "__main__":
    # prepare arithmetic coding

    # the example message "AABBBBBBBC" leads to the frequency 
    # dictionary {'A': 0.2, 'B': 0.7, 'C': 0.1} as used in the lecture
    alphabeth_with_freq = "AABBBBBBBC"
    
    # get the frequencies of the characters
    frequencies = get_character_counts(alphabeth_with_freq)
    print("Character Frequencies:", frequencies)
    # calculate the total number of characters
    total = sum(frequencies.values())
    # calculate probabilities
    for char, freq in frequencies.items():
        frequencies[char] = freq / total
    print("Character probabilities:", frequencies) # sorted by frequency
    # sort lexicographically
    frequencies = dict(sorted(frequencies.items(), key=lambda item: item[0], reverse=False))
    print("Character probabilities (sorted):", frequencies)
    # calculate the cumulative frequencies
    cumulative_frequencies = {}
    cumulative_frequency = 0
    for char, freq in frequencies.items():
        cumulative_frequencies[char] = cumulative_frequency
        cumulative_frequency += freq

    # run arithmetic coding
    original_msg ="ABC"
    print("Original message:", original_msg)
    arithmetic_code = encode(original_msg, frequencies, cumulative_frequencies)
    print("Arithmetic Code:", arithmetic_code)
    #print("Encoded message:", get_code_from_interval(arithmetic_code))
    #print("Encoded message length:", len(get_code_from_interval(arithmetic_code)))

    # test decoding
    decoded_msg = decode_with_length(arithmetic_code, frequencies, cumulative_frequencies, len(original_msg))
    print("Decoded message:", decoded_msg)
    # if the decoding fails, the precision might be the issue
    print("Message Decoded Successfully? {result}".format(result=original_msg == decoded_msg))
